package parser;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.StringTokenizer;

public class CFGConvert {

	private static void usage() {
		System.out.println("java CFGConvert <file>");
	}

	private final HashSet<Rule> rules;
	private final NameFactory nf;
	private final HashMap<String, ArrayList<Rule>> nonterminals;

	public CFGConvert() {
		nf = new NameFactory();
		rules = new HashSet<Rule>();
		nonterminals = new HashMap<String, ArrayList<Rule>>();
	}

	public void addRule(String s, int line) throws ParseException {
		if (!s.isEmpty()) {
			Rule r = new Rule(s, line);
			rules.add(r);
		}
	}

	public String toString() {
		StringBuffer b = new StringBuffer();
		for (Rule r : rules) {
			b.append(r);
			b.append("\n");
		}
		return b.toString();
	}

	private void registerSymbols() {
		// register the symbols
		for (Rule r : rules) {
			ArrayList<Rule> ntRules = nonterminals.get(r.getLHS());
			if (ntRules == null) {
				ntRules = new ArrayList<Rule>();
				nonterminals.put(r.getLHS(), ntRules);
			}
			ntRules.add(r);
		}
	}

	private boolean isCNF() {
		for (Rule r : rules) {
			if (r.isBinary() || r.isTerminal()) {
			} else {
				return false;
			}
		}
		return true;
	}

	private void convertToCNF() {
		// get rid of the n-ples
		Iterator<Rule> ruleIter = rules.iterator();

		ArrayList<Rule> newRules = new ArrayList<Rule>();
		while (ruleIter.hasNext()) {
			Rule nextRule = ruleIter.next();
			if (nextRule.threeOrMore()) {
				ruleIter.remove();
				nextRule.reduceOrder(newRules, nf);
			}
		}
		for (Rule r : newRules) {
			rules.add(r);
		}

		newRules.clear();
		for (Rule r : rules) {
			if (r.isBinary() || r.isTerminal() || r.isUnary()) {

			} else {
				throw new RuntimeException("invalid rule: " + r);
			}
		}

		// now the symbols are set in stone
		registerSymbols();
		// find the unary rules, and get rid of them by promoting the RHS
		// directly to the LHS.
		removeUnaryRules();
	}

	private void removeUnaryRules() {
		Iterator<Rule> ruleIter;
		ArrayList<Rule> newRules = new ArrayList<Rule>();
		while (!isCNF()) {
			ruleIter = rules.iterator();
			while (ruleIter.hasNext()) {
				Rule nextRule = ruleIter.next();
				if (nextRule.isUnary()) {
					ruleIter.remove();
					ArrayList<Rule> copyRules = nonterminals.get(nextRule
							.getRHS().get(0));
					if (copyRules == null)
						throw new RuntimeException("couldn't find a rule: "
								+ nextRule);
					for (Rule c : copyRules) {
						LinkedList<String> l = new LinkedList<String>();
						for (String s : c.getRHS()) {
							l.add(s);
						}
						Rule newRule = new Rule(nextRule.getLHS(), l,
								nextRule.getProbability());
						newRules.add(newRule);
					}
				}
			}
			for (Rule r : newRules) {
				rules.add(r);
			}
			newRules.clear();
		}
	}

	public static void main(String[] args) throws IOException, ParseException {
		if (args.length != 1) {
			usage();
			return;
		}

		CFGConvert c = new CFGConvert();

		DataInputStream in;
		FileInputStream fstream = null;
		try {
			fstream = new FileInputStream(args[0]);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		if (fstream == null) {
			System.err.println("failed to open file");
			return;
		}
		// Get the object of DataInputStream
		in = new DataInputStream(fstream);
		BufferedReader br = new BufferedReader(new InputStreamReader(in));
		String strLine;
		// Read File Line By Line
		int line = 0;
		while ((strLine = br.readLine()) != null) {
			c.addRule(strLine, line);
		}
		in.close();

		c.convertToCNF();

		String fname = args[0];
		fname = (new StringTokenizer(fname, ".")).nextToken();

		// Create file
		FileWriter ofstream = new FileWriter(fname + ".cnf");
		BufferedWriter out = new BufferedWriter(ofstream);
		out.write(c.toString());
		out.close();

	}

}
