package parser;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.LinkedList;

public class Rule {
	private final String lhs;
	private final LinkedList<String> rhs;
	private final double probability;
	
	public Rule(String s, int line) throws ParseException {
		String[] r = s.split("\\s+");
		if (r.length < 2) {
			throw new ParseException("Rule " + s + " too small", line);
		}
		lhs = r[0];
		rhs = new LinkedList<String>();
		for (int i = 1; i < r.length - 1; i++) {
			rhs.add(r[i]);
		}
		double tempProb = 1.0;
		try{
			tempProb = Double.parseDouble(r[r.length - 1]);
		} catch (NumberFormatException nfe){
			System.err.println("Warning: no probability for this rule: <" + s + "> using default probability of 1.0");
			rhs.add(r[r.length - 1]);
		}
		probability = tempProb;
		validateRule(s, line);
	}

	public String getLHS() {
		return lhs;
	}

	public LinkedList<String> getRHS() {
		return rhs;
	}

	public boolean lhsMatches(String s) {
		return s.equals(lhs);
	}
	
	public boolean rhsMatches(String s){
		return rhs.contains(s);
	}

	public static boolean acceptableSymbol(String s) {
		if (s.matches("^[a-zA-Z]+$"))
			return true;
		else
			return false;
	}
	
	Rule(String l, LinkedList<String> r, double p){
		assert(allUpperCase(l));
		this.lhs = l;
		this.rhs = r;
		this.probability = p;
	}
	
	public boolean isBinary(){
		return rhs.size() == 2;
	}
	
	public double getProbability(){
		return probability;
	}
	
	public boolean isUnary(){
		return rhs.size() == 1 && allUpperCase(rhs.get(0));
	}
	
	public boolean isTerminal(){
		return rhs.size() == 1 && allLowerCase(rhs.get(0));
	}
	
	public void reduceOrder(ArrayList<Rule> newRules, NameFactory nf){
		if(rhs.size() == 2){
			newRules.add(this);
		}
		else {
			String newSymbol = nf.nextName();
			LinkedList<String> binaryRHS = new LinkedList<String>();
			binaryRHS.add(new String(rhs.get(0)));
			binaryRHS.add(newSymbol);
			Rule newBinary = new Rule(new String(lhs), binaryRHS, 1.0);
			newRules.add(newBinary);
			rhs.remove(0);
			
			Rule remainingRule = new Rule(new String(newSymbol), rhs, probability);
			remainingRule.reduceOrder(newRules, nf);
		}
	}
	
	public boolean threeOrMore(){
		return rhs.size() > 2;
	}
	
	public static boolean allUpperCase(String s) {
		return s.toUpperCase().equals(s);
	}

	public static boolean allLowerCase(String s) {
		return s.toLowerCase().equals(s);
	}

	private void validateRule(String s, int line) throws ParseException {
		if (!allUpperCase(lhs)) {
			throw new ParseException(s
					+ " has a non uppercase left hand side - ", line);
		}
		if (!acceptableSymbol(lhs)) {
			throw new ParseException("<" + lhs + "> has unacceptable characters - " + line,
					line);
		}

		for (String symbol : rhs) {
			if (allUpperCase(symbol) || allLowerCase(symbol)) {

			} else {
				throw new ParseException(
						symbol
								+ " is mixed case, not clear if it is a terminal or a nonterminal - ",
						line);
			}
			if (!acceptableSymbol(symbol)) {
				throw new ParseException(s + " has unacceptable characters - ",
						line);
			}
		}
	}
	
	public String toString(){
		StringBuffer b = new StringBuffer();
		
		b.append(lhs);
		b.append(" ");
		for(String r : rhs){
			b.append(r);
			b.append(" ");
		}
		b.append(probability);
		return b.toString();
	}
}
